# distutils: language = c++
# distutils: libraries = capnpc capnp capnp-rpc
# distutils: sources = pytact/graph_api.capnp.cpp
# cython: c_string_type = str
# cython: c_string_encoding = default
# cython: embedsignature = True
# cython: language_level = 3

import os
import capnp
import pytact.graph_api_capnp

cdef class Graph_Node_Reader:
    """
    Fits exactly in 128 bits.
    A node has a label that optionally contains some additional information, together with a list of
    outgoing edges (children).
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_Node()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Graph_Node_Reader source, object root):
        cdef Graph_Node_Reader wrapper = Graph_Node_Reader.__new__(Graph_Node_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def label(self):
        """
        Inlined for efficiency purposes
        Proof state
        Hash a unique id (evar) for the proof state that distinquishes proof states with identical
        contents but do not point to the same object nonetheless
        """
        return Graph_Node_Label_Reader.init(self.source.getLabel(), self.root)

    @property
    def children_index(self):
        return self.source.getChildrenIndex()

    @property
    def children_count(self):
        """
        The children of a node are encoded as a range withing the `edges`-list of the graph.
        """
        return self.source.getChildrenCount()

    @property
    def identity(self):
        """
        The identity of a node uniquely determines it. That is, one can consider any to nodes with the same
        identity to be equal. The notion of equal we use is as follows:
        1. Graph perspective: Two nodes have the same identity if and only if they are bisimilar.
           In this notion, bisimilarity does take into consideration the label of the nodes, modulo some
           equivalence class that is not fully specified here. One aspect of the equivalence class is that
           for definition nodes their associated global context (accessed through `Definition.previous`) is
           not taken into account.
        2. Lambda calculus perspective: Two nodes have the same identity if their corresponding lambda terms
           are alpha-equivalent. Note that two definitions with the same body are not considered
           alpha-equivalent.

        The identity of a node is used to perform partial graph-sharing. That is, two nodes with the same
        identity are merged when the graph is generated. There are three reasons why two nodes with the same
        semantic identity might have a different physical identity:
        1. Nodes are only merged when the first node exists in the same graph as the second node, or exists
           in a dependent graph. Hence, nodes originating from developments that do not depend on each other
           are never merged. Full graph-sharing would require global analysis on a dataset, which any consumer
           can optionally do as a post-processing step.
        2. Two definition nodes with the same name and body have the same identity. But if they occur in
           different global contexts, these nodes are physically different to ensure the integrity of their
           global contexts.
        3. For definitions with an opaque proof, the sub-graph that represents this opaque proof is ignored
           while calculating the identity. Morally we defend this decision with the concept that opaque proofs
           are invisible in Coq (proof irrelevance) and two constants with different proofs might as well be
           considered equal. Practically speaking, this is done so that one can compare the identity of a
           definition in a dataset quickly to a identity of a live definition during Coq interaction without
           having to produce the entire graph of opaque proofs (which tend to be large).

        Beware that the identity is currently a 64 bit field. For datasets that have graphs of size in the
        order of billions of nodes there is a non-trivial chance of a collision. (We consider this acceptable
        for machine learning purposes.)
        """
        return self.source.getIdentity()
pytact.graph_api_capnp.Graph.Node.schema = _Schema()._init(getGraph_Node_Schema()).as_struct()

cdef class Graph_EdgeTarget_Reader:
    """
    Fits exactly in 64 bits. Let's keep it that way.
    The 'pointy' end of an edge, together with the label of that edge.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_EdgeTarget()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Graph_EdgeTarget_Reader source, object root):
        cdef Graph_EdgeTarget_Reader wrapper = Graph_EdgeTarget_Reader.__new__(Graph_EdgeTarget_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def label(self):
        return None(self.source.getLabel())

    @property
    def target(self):
        return Graph_EdgeTarget_Target_Reader.init(self.source.getTarget(), self.root)
pytact.graph_api_capnp.Graph.EdgeTarget.schema = _Schema()._init(getGraph_EdgeTarget_Schema()).as_struct()

cdef class Graph_Reader:
    """
    A graph is the main data store (the 'heap') that contains the bulk of the data in the dataset and
    during communication with Coq. A graph is a collection of labeled nodes with directed, labeled edges
    between them. A graph may reference nodes from other graphs. For an edge 'A --> B' in the graph it
    always holds that 'A' is part of the current graph, but 'B' may (or may not) be part of another graph.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Graph_Reader source, object root):
        cdef Graph_Reader wrapper = Graph_Reader.__new__(Graph_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def nodes(self):
        return Graph_Node_Reader_List.init(self.source.getNodes(), self.root)
    @property
    def has_nodes(self):
        return self.source.hasNodes()

    @property
    def edges(self):
        """
        The main memory store of the graph. It acts as a heap similar to the main memory of a C/C++ program.
        The heap is accessed by indexing the `nodes` list using a `NodeIndex` which returns a `Node`.
        Every node has a label and a list of children, which is indicated as a range within the `edges` list using
        `childrenIndex` and `childrenCount`. The targets of the edges can again be found in the `nodes` list of the
        current file or of a dependency.
        Note that just like in C/C++ doing pointer arithmetic on the heap is undefined behavior, and you may
        encounter arbitrary garbage if you do this. In particular, iterating over the heap is discouraged.
        Instead, you should access the heap through various entry-points that are provided.
        """
        return Graph_EdgeTarget_Reader_List.init(self.source.getEdges(), self.root)
    @property
    def has_edges(self):
        return self.source.hasEdges()
pytact.graph_api_capnp.Graph.schema = _Schema()._init(getGraph_Schema()).as_struct()

cdef class IntP_Reader:

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asIntP()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_IntP_Reader source, object root):
        cdef IntP_Reader wrapper = IntP_Reader.__new__(IntP_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def value(self):
        return self.source.getValue()
pytact.graph_api_capnp.IntP.schema = _Schema()._init(getIntP_Schema()).as_struct()

cdef class Definition_Reader:
    """
    A definition of the CIC, which is either an constant, inductive, constructor, projection or section
    variable. Constants and section variables can have tactical proofs associated to them.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Definition_Reader source, object root):
        cdef Definition_Reader wrapper = Definition_Reader.__new__(Definition_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
    @property
    def which(self):
        return Definition_Which(self.source.which())
    @property
    def which_raw(self):
        return self.source.which()

    @property
    def name(self):
        """
        The fully-qualified name of the definition. The name should be unique in a particular global context,
        but is not unique among different branches of the global in a dataset.
        """
        temp = self.source.getName()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_name(self):
        return self.source.hasName()

    @property
    def previous(self):
        """
        The previous definition within the global context of the current file.
        For the first definition this field is set to `len(graph.nodes)`.
        The contract on this field is that any definition nodes reachable from the forward closure of the definition
        must also be reachable through the chain of previous fields. An exception to this rule are mutually
        recursive definitions. Those nodes are placed into the global context in an arbitrary ordering.
        """
        return self.source.getPrevious()

    @property
    def external_previous(self):
        """
        This field provides information about the external global context.
        At any point in a source file other files 'X' can be loaded through 'Require X'. When this happens, the
        definitions of X that are reachable through its 'representative' field become available to all subsequent
        definitions.
        """
        return Uint16_List.init(self.source.getExternalPrevious(), self.root)
    @property
    def has_external_previous(self):
        return self.source.hasExternalPrevious()

    @property
    def status(self):
        """
        A definition can have different origins, which are encoded here
        """
        return Definition_Status_Reader.init(self.source.getStatus(), self.root)

    @property
    def inductive(self):
        return self.source.getInductive()
    @property
    def is_inductive(self):
        return self.source.isInductive()

    @property
    def constructor(self):
        return self.source.getConstructor()
    @property
    def is_constructor(self):
        return self.source.isConstructor()

    @property
    def projection(self):
        """
        These definitions are part of a mutually recursive cluster. They hold a reference to another definition
        that acts as the 'representative' of the mutually recursive cluster. The representative is chosen such
        that all definitions of the cluster are reachable through its `previous` pointer. Additionally, all
        definitions within the cluster have the same representative, and no definitions that are not part of the
        cluster are interleaved within the chain of `previous` nodes.
        """
        return self.source.getProjection()
    @property
    def is_projection(self):
        return self.source.isProjection()

    @property
    def manual_constant(self):
        """
        A constant defined by directly inputting a term
        In the future, we might augment such constants with tactical
        refinement proofs that build the term iteratively.
        """
        return self.source.getManualConstant()
    @property
    def is_manual_constant(self):
        return self.source.isManualConstant()

    @property
    def tactical_constant(self):
        """
        A constant that was either directly or indirectly generated by a tactical proof.
        Note that for non-original constants, the proof step sequence may not be very accurate.
        """
        return ProofStep_Reader_List.init(self.source.getTacticalConstant(), self.root)
    @property
    def is_tactical_constant(self):
        return self.source.isTacticalConstant()
    @property
    def has_tactical_constant(self):
        return self.source.hasTacticalConstant()

    @property
    def manual_section_constant(self):
        """
        A section variable or local section definition.
        """
        return self.source.getManualSectionConstant()
    @property
    def is_manual_section_constant(self):
        return self.source.isManualSectionConstant()

    @property
    def tactical_section_constant(self):
        """
        Local section definitions can also be defined using a tactical proof.
        """
        return ProofStep_Reader_List.init(self.source.getTacticalSectionConstant(), self.root)
    @property
    def is_tactical_section_constant(self):
        return self.source.isTacticalSectionConstant()
    @property
    def has_tactical_section_constant(self):
        return self.source.hasTacticalSectionConstant()

    @property
    def type_text(self):
        """
        A textual representation of the type of this definition.
        """
        temp = self.source.getTypeText()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_type_text(self):
        return self.source.hasTypeText()

    @property
    def term_text(self):
        """
        A textual representation of the body of this definition.
        For inductives, constructors, projections, section variables and axioms the string is empty.
        """
        temp = self.source.getTermText()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_term_text(self):
        return self.source.hasTermText()
pytact.graph_api_capnp.Definition.schema = _Schema()._init(getDefinition_Schema()).as_struct()

cdef class FloatP_Reader:

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asFloatP()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_FloatP_Reader source, object root):
        cdef FloatP_Reader wrapper = FloatP_Reader.__new__(FloatP_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def value(self):
        return self.source.getValue()
pytact.graph_api_capnp.FloatP.schema = _Schema()._init(getFloatP_Schema()).as_struct()

cdef class Graph_Node_Label_Reader:
    """
    Inlined for efficiency purposes
    Proof state
    Hash a unique id (evar) for the proof state that distinquishes proof states with identical
    contents but do not point to the same object nonetheless
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_Node_Label()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Graph_Node_Label_Reader source, object root):
        cdef Graph_Node_Label_Reader wrapper = Graph_Node_Label_Reader.__new__(Graph_Node_Label_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
    @property
    def which(self):
        return Graph_Node_Label_Which(self.source.which())
    @property
    def which_raw(self):
        return self.source.which()

    @property
    def proof_state(self):
        return IntP_Reader.init(self.source.getProofState(), self.root)
    @property
    def is_proof_state(self):
        return self.source.isProofState()
    @property
    def has_proof_state(self):
        return self.source.hasProofState()

    @property
    def context_def(self):
        return self.source.getContextDef()
    @property
    def is_context_def(self):
        return self.source.isContextDef()

    @property
    def context_assum(self):
        return self.source.getContextAssum()
    @property
    def is_context_assum(self):
        return self.source.isContextAssum()

    @property
    def definition(self):
        return Definition_Reader.init(self.source.getDefinition(), self.root)
    @property
    def is_definition(self):
        return self.source.isDefinition()
    @property
    def has_definition(self):
        return self.source.hasDefinition()

    @property
    def const_empty(self):
        return self.source.getConstEmpty()
    @property
    def is_const_empty(self):
        return self.source.isConstEmpty()

    @property
    def sort_s_prop(self):
        return self.source.getSortSProp()
    @property
    def is_sort_s_prop(self):
        return self.source.isSortSProp()

    @property
    def sort_prop(self):
        return self.source.getSortProp()
    @property
    def is_sort_prop(self):
        return self.source.isSortProp()

    @property
    def sort_set(self):
        return self.source.getSortSet()
    @property
    def is_sort_set(self):
        return self.source.isSortSet()

    @property
    def sort_type(self):
        """
        Collapsed universe
        """
        return self.source.getSortType()
    @property
    def is_sort_type(self):
        return self.source.isSortType()

    @property
    def rel(self):
        return self.source.getRel()
    @property
    def is_rel(self):
        return self.source.isRel()

    @property
    def evar(self):
        return self.source.getEvar()
    @property
    def is_evar(self):
        return self.source.isEvar()

    @property
    def evar_subst(self):
        return self.source.getEvarSubst()
    @property
    def is_evar_subst(self):
        return self.source.isEvarSubst()

    @property
    def cast(self):
        return self.source.getCast()
    @property
    def is_cast(self):
        return self.source.isCast()

    @property
    def prod(self):
        return self.source.getProd()
    @property
    def is_prod(self):
        return self.source.isProd()

    @property
    def lambda_(self):
        return self.source.getLambda()
    @property
    def is_lambda_(self):
        return self.source.isLambda()

    @property
    def let_in(self):
        return self.source.getLetIn()
    @property
    def is_let_in(self):
        return self.source.isLetIn()

    @property
    def app(self):
        return self.source.getApp()
    @property
    def is_app(self):
        return self.source.isApp()

    @property
    def case(self):
        return self.source.getCase()
    @property
    def is_case(self):
        return self.source.isCase()

    @property
    def case_branch(self):
        return self.source.getCaseBranch()
    @property
    def is_case_branch(self):
        return self.source.isCaseBranch()

    @property
    def fix(self):
        return self.source.getFix()
    @property
    def is_fix(self):
        return self.source.isFix()

    @property
    def fix_fun(self):
        return self.source.getFixFun()
    @property
    def is_fix_fun(self):
        return self.source.isFixFun()

    @property
    def co_fix(self):
        return self.source.getCoFix()
    @property
    def is_co_fix(self):
        return self.source.isCoFix()

    @property
    def co_fix_fun(self):
        return self.source.getCoFixFun()
    @property
    def is_co_fix_fun(self):
        return self.source.isCoFixFun()

    @property
    def int(self):
        return IntP_Reader.init(self.source.getInt(), self.root)
    @property
    def is_int(self):
        return self.source.isInt()
    @property
    def has_int(self):
        return self.source.hasInt()

    @property
    def float(self):
        return FloatP_Reader.init(self.source.getFloat(), self.root)
    @property
    def is_float(self):
        return self.source.isFloat()
    @property
    def has_float(self):
        return self.source.hasFloat()

    @property
    def primitive(self):
        temp = self.source.getPrimitive()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def is_primitive(self):
        return self.source.isPrimitive()
    @property
    def has_primitive(self):
        return self.source.hasPrimitive()

cdef class Definition_Status_Substituted_Reader:
    """
    A definition that was obtained by performing some sort of module functor substitution.
    We reference the original definition.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition_Status_Substituted()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Definition_Status_Substituted_Reader source, object root):
        cdef Definition_Status_Substituted_Reader wrapper = Definition_Status_Substituted_Reader.__new__(Definition_Status_Substituted_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def dep_index(self):
        """
        Indicates to which graph a node belongs. How this should be resolved is not specified here.
        However, a value of zero always points to the 'current' graph.
        """
        return self.source.getDepIndex()

    @property
    def node_index(self):
        """
        The index into `Graph.nodes` where this node can be found.
        """
        return self.source.getNodeIndex()

cdef class Definition_Status_Reader:
    """
    A definition can have different origins, which are encoded here
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDefinition_Status()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Definition_Status_Reader source, object root):
        cdef Definition_Status_Reader wrapper = Definition_Status_Reader.__new__(Definition_Status_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
    @property
    def which(self):
        return Definition_Status_Which(self.source.which())
    @property
    def which_raw(self):
        return self.source.which()

    @property
    def original(self):
        """
        An object as originally inputted by the user.
        """
        return self.source.getOriginal()
    @property
    def is_original(self):
        return self.source.isOriginal()

    @property
    def discharged(self):
        """
        a definition that was originally defined in a section and has now had the section
        variables discharged into it. The node index references the original definition.
        """
        return self.source.getDischarged()
    @property
    def is_discharged(self):
        return self.source.isDischarged()

    @property
    def substituted(self):
        """
        A definition that was obtained by performing some sort of module functor substitution.
        We reference the original definition.
        """
        return Definition_Status_Substituted_Reader.init(self.source.getSubstituted(), self.root)
    @property
    def is_substituted(self):
        return self.source.isSubstituted()

cdef class ProofStep_Reader:
    """
    A proof step is the execution of a single tactic on one or more proof states, producing a list of outcomes.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofStep()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_ProofStep_Reader source, object root):
        cdef ProofStep_Reader wrapper = ProofStep_Reader.__new__(ProofStep_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def tactic(self):
        """
        The tactic that generated the proof step. Note that the arguments of the tactic can be found in the
        individual outcomes, because they may be different for each outcome.
        """
        return ProofStep_Tactic_Reader.init(self.source.getTactic(), self.root)

    @property
    def outcomes(self):
        """
        A list of transformations of proof states to other proof states, as executed by the tactic of the proof step
        """
        return Outcome_Reader_List.init(self.source.getOutcomes(), self.root)
    @property
    def has_outcomes(self):
        return self.source.hasOutcomes()
pytact.graph_api_capnp.ProofStep.schema = _Schema()._init(getProofStep_Schema()).as_struct()

cdef class Tactic_Reader:
    """
    A concrete tactic with it's parameters determined. Somewhat strangely, this struct does not actually include
    these parameters. They can instead be found in `Outcome.tacticArguments`. The reason for this is that one
    tactic can run on multiple proof states at the same time and for all of those proof states, the arguments
    may be resolved differently.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asTactic()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Tactic_Reader source, object root):
        cdef Tactic_Reader wrapper = Tactic_Reader.__new__(Tactic_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def ident(self):
        """
        A hash representing the identity of a tactic without it's arguments. Due to the complexity of the syntax
        trees of Coq's tactics, we do not currently encode the syntax tree. Instead, this hash is a representative
        of the syntax tree of the tactic with all of it's arguments removed.
        """
        return self.source.getIdent()

    @property
    def text(self):
        """
        The full text of the tactic including the full arguments. This does not currently correspond to
        (ident, arguments) because in this dataset arguments do not include full terms, but only references to
        definitions and local context elements. Tactics are postprocessed to remove explicit naming of new
        hypotheses as much as possible, and instead asking Coq to invent a name.
        """
        temp = self.source.getText()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_text(self):
        return self.source.hasText()

    @property
    def base_text(self):
        """
        A textual representation of the base tactic without arguments. It tries to roughly correspond to `ident`.
        Note, however, that this is both an under-approximation and an over-approximation. The reason is that tactic
        printing is not 100% isomorphic to Coq's internal AST of tactics. Sometimes, different tactics get mapped to
        the same text. Conversely, the same tactic may be mapped to different texts when identifiers are printed
        using different partially-qualified names.
        """
        temp = self.source.getBaseText()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_base_text(self):
        return self.source.hasBaseText()

    @property
    def interm_text(self):
        """
        A textual representation that tries to come as close as possible to (ident, arguments).
        It comes with the same caveats as `baseText`.
        """
        temp = self.source.getIntermText()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_interm_text(self):
        return self.source.hasIntermText()

    @property
    def exact(self):
        """
        Indicates whether or not `ident` + `arguments` is faithfully reversible into the original "strictified" tactic.
        Note that this does not necessarily mean that it represents exactly the tactic that was inputted by the user.
        All tactics are modified to be 'strict' (meaning that tactics that have delayed variables in them break).
        This flag measures the faithfulness of the representation w.r.t. the strict version of the tactic, not the
        original tactic inputted by the user.
        """
        return self.source.getExact()

    @property
    def text_non_anonymous(self):
        """
        Same as `text` except that the tactic is not postprocessed to remove explicit naming of new hypotheses.
        """
        temp = self.source.getTextNonAnonymous()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_text_non_anonymous(self):
        return self.source.hasTextNonAnonymous()
pytact.graph_api_capnp.Tactic.schema = _Schema()._init(getTactic_Schema()).as_struct()

cdef class ProofStep_Tactic_Reader:
    """
    The tactic that generated the proof step. Note that the arguments of the tactic can be found in the
    individual outcomes, because they may be different for each outcome.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofStep_Tactic()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_ProofStep_Tactic_Reader source, object root):
        cdef ProofStep_Tactic_Reader wrapper = ProofStep_Tactic_Reader.__new__(ProofStep_Tactic_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
    @property
    def which(self):
        return ProofStep_Tactic_Which(self.source.which())
    @property
    def which_raw(self):
        return self.source.which()

    @property
    def unknown(self):
        """
        Sometimes a tactic cannot or should not be recorded. In those cases, it is marked as 'unknown'.
        This currently happens with tactics that are run as a result of the `Proof with tac` construct and it
        happens for tactics that are known to be unsafe like `change_no_check`, `fix`, `cofix` and more.
        """
        return self.source.getUnknown()
    @property
    def is_unknown(self):
        return self.source.isUnknown()

    @property
    def known(self):
        """
        The tactic
        """
        return Tactic_Reader.init(self.source.getKnown(), self.root)
    @property
    def is_known(self):
        return self.source.isKnown()
    @property
    def has_known(self):
        return self.source.hasKnown()

cdef class Outcome_Reader:
    """
    An outcome is the result of running a tactic on a proof state. A tactic may run on multiple proof states.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asOutcome()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Outcome_Reader source, object root):
        cdef Outcome_Reader wrapper = Outcome_Reader.__new__(Outcome_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def before(self):
        """
        The proof state before the tactic execution.
        """
        return ProofState_Reader.init(self.source.getBefore(), self.root)
    @property
    def has_before(self):
        return self.source.hasBefore()

    @property
    def after(self):
        """
        The new proof states that were generated by the tactic.
        """
        return ProofState_Reader_List.init(self.source.getAfter(), self.root)
    @property
    def has_after(self):
        return self.source.hasAfter()

    @property
    def term(self):
        """
        The proof term that witnesses the transition from the before state to the after states. It contains a hole
        (an `evar` node) for each of the after states. It may also refer to elements of the local context of the
        before state.
        """
        return Outcome_Term_Reader.init(self.source.getTerm(), self.root)

    @property
    def term_text(self):
        """
        A textual representation of the proof term.
        """
        temp = self.source.getTermText()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_term_text(self):
        return self.source.hasTermText()

    @property
    def tactic_arguments(self):
        """
        The arguments of the tactic that produced this outcome. Note that these arguments belong to the tactic in
        `ProofStep.tactic`.
        """
        return Argument_Reader_List.init(self.source.getTacticArguments(), self.root)
    @property
    def has_tactic_arguments(self):
        return self.source.hasTacticArguments()
pytact.graph_api_capnp.Outcome.schema = _Schema()._init(getOutcome_Schema()).as_struct()

cdef class Outcome_Term_Reader:
    """
    The proof term that witnesses the transition from the before state to the after states. It contains a hole
    (an `evar` node) for each of the after states. It may also refer to elements of the local context of the
    before state.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asOutcome_Term()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Outcome_Term_Reader source, object root):
        cdef Outcome_Term_Reader wrapper = Outcome_Term_Reader.__new__(Outcome_Term_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def dep_index(self):
        """
        Indicates to which graph a node belongs. How this should be resolved is not specified here.
        However, a value of zero always points to the 'current' graph.
        """
        return self.source.getDepIndex()

    @property
    def node_index(self):
        """
        The index into `Graph.nodes` where this node can be found.
        """
        return self.source.getNodeIndex()

cdef class ProofState_Reader:
    """
    A proof state represents a particular point in the tactical proof of a constant.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofState()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_ProofState_Reader source, object root):
        cdef ProofState_Reader wrapper = ProofState_Reader.__new__(ProofState_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def root(self):
        """
        The entry-point of the proof state, all nodes that are 'part of' the proof state are reachable from here.
        """
        return ProofState_Root_Reader.init(self.source.getRoot(), self.root)

    @property
    def context(self):
        """
        The local context of the proof state. These nodes label's are either `contextAssum` or `contextDef`. Note that
        these nodes are also reachable from the root of the proof state.
        """
        return Node_Reader_List.init(self.source.getContext(), self.root)
    @property
    def has_context(self):
        return self.source.hasContext()

    @property
    def context_names(self):
        """
        The names of the local context nodes of the proof state, as they originally appeared in the proof.
        These names should be used for debugging and viewing purposes only, because hypothesis-generating tactics have
        been modified to use auto-generated names. Hence, tactics should not be concerned about the names of
        the context.
        """
        return String_List.init(self.source.getContextNames(), self.root)
    @property
    def has_context_names(self):
        return self.source.hasContextNames()

    @property
    def context_text(self):
        """
        A textual representation of the type/definition of context nodes
        """
        return String_List.init(self.source.getContextText(), self.root)
    @property
    def has_context_text(self):
        return self.source.hasContextText()

    @property
    def conclusion_text(self):
        """
        A textual representation of the conclusion of the proof state.
        """
        temp = self.source.getConclusionText()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_conclusion_text(self):
        return self.source.hasConclusionText()

    @property
    def text(self):
        """
        A textual representation of the proof state.
        """
        temp = self.source.getText()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_text(self):
        return self.source.hasText()

    @property
    def id(self):
        """
        A unique identifier of the proof state. Any two proof states in a tactical proof that have an equal id
        can morally be regarded to be 'the same' proof state.
        IMPORTANT: Two proof states with the same id may still have different contents. This is because proof states
                   can contain existential variables (represented by the `evar` node) that can be filled as a
                   side-effect by a tactic running on another proof state.
        TODO: This is currently duplicated with the id attached to the `ProofState` node that `root` points to.
              Eventually, this id should be resolved to a NodeIndex that represents the location of this proof state
              in the final proof term that was generated.
        """
        return self.source.getId()
pytact.graph_api_capnp.ProofState.schema = _Schema()._init(getProofState_Schema()).as_struct()

cdef class Argument_Reader:
    """
    A concrete argument of a tactic.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asArgument()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Argument_Reader source, object root):
        cdef Argument_Reader wrapper = Argument_Reader.__new__(Argument_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
    @property
    def which(self):
        return Argument_Which(self.source.which())
    @property
    def which_raw(self):
        return self.source.which()

    @property
    def unresolvable(self):
        """
        An argument that is currently unresolvable due to limitations of the extraction process.
        """
        return self.source.getUnresolvable()
    @property
    def is_unresolvable(self):
        return self.source.isUnresolvable()

    @property
    def term(self):
        """
        The root of a graph representing an argument that is a term in the calculus of constructions.
        """
        return Argument_Term_Reader.init(self.source.getTerm(), self.root)
    @property
    def is_term(self):
        return self.source.isTerm()
pytact.graph_api_capnp.Argument.schema = _Schema()._init(getArgument_Schema()).as_struct()

cdef class ProofState_Root_Reader:
    """
    The entry-point of the proof state, all nodes that are 'part of' the proof state are reachable from here.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asProofState_Root()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_ProofState_Root_Reader source, object root):
        cdef ProofState_Root_Reader wrapper = ProofState_Root_Reader.__new__(ProofState_Root_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def dep_index(self):
        """
        Indicates to which graph a node belongs. How this should be resolved is not specified here.
        However, a value of zero always points to the 'current' graph.
        """
        return self.source.getDepIndex()

    @property
    def node_index(self):
        """
        The index into `Graph.nodes` where this node can be found.
        """
        return self.source.getNodeIndex()

cdef class Node_Reader:
    """
    A node in the CIC graph
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asNode()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Node_Reader source, object root):
        cdef Node_Reader wrapper = Node_Reader.__new__(Node_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def dep_index(self):
        """
        Indicates to which graph a node belongs. How this should be resolved is not specified here.
        However, a value of zero always points to the 'current' graph.
        """
        return self.source.getDepIndex()

    @property
    def node_index(self):
        """
        The index into `Graph.nodes` where this node can be found.
        """
        return self.source.getNodeIndex()
pytact.graph_api_capnp.Node.schema = _Schema()._init(getNode_Schema()).as_struct()

cdef class Argument_Term_Reader:
    """
    The root of a graph representing an argument that is a term in the calculus of constructions.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asArgument_Term()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Argument_Term_Reader source, object root):
        cdef Argument_Term_Reader wrapper = Argument_Term_Reader.__new__(Argument_Term_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def dep_index(self):
        """
        Indicates to which graph a node belongs. How this should be resolved is not specified here.
        However, a value of zero always points to the 'current' graph.
        """
        return self.source.getDepIndex()

    @property
    def node_index(self):
        """
        The index into `Graph.nodes` where this node can be found.
        """
        return self.source.getNodeIndex()

cdef class Graph_EdgeTarget_Target_Reader:

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asGraph_EdgeTarget_Target()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Graph_EdgeTarget_Target_Reader source, object root):
        cdef Graph_EdgeTarget_Target_Reader wrapper = Graph_EdgeTarget_Target_Reader.__new__(Graph_EdgeTarget_Target_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def dep_index(self):
        """
        Indicates to which graph a node belongs. How this should be resolved is not specified here.
        However, a value of zero always points to the 'current' graph.
        """
        return self.source.getDepIndex()

    @property
    def node_index(self):
        """
        The index into `Graph.nodes` where this node can be found.
        """
        return self.source.getNodeIndex()

cdef class AbstractTactic_Reader:

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asAbstractTactic()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_AbstractTactic_Reader source, object root):
        cdef AbstractTactic_Reader wrapper = AbstractTactic_Reader.__new__(AbstractTactic_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def ident(self):
        """
        An abstract tactic is referenced to using a identifier (hash).
        """
        return self.source.getIdent()

    @property
    def parameters(self):
        """
        Every tactic has a constant number of parameters that need to be filled in.
        """
        return self.source.getParameters()
pytact.graph_api_capnp.AbstractTactic.schema = _Schema()._init(getAbstractTactic_Schema()).as_struct()

cdef class DataVersion_Reader:
    """
    Version info of a dataset.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDataVersion()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_DataVersion_Reader source, object root):
        cdef DataVersion_Reader wrapper = DataVersion_Reader.__new__(DataVersion_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def major(self):
        """
        Currently we only have a major version. Any change to the graph format, Cap'n Proto schema or
        communication protocol is considered a breaking change and will increment the major version number.
        Note that currently we don't distinquish backwards-compatible Cap'n Proto schema changes from
        non-backwards-compatible changes. The schema might change in a non-compatible way without notice.
        In the future, a minor version id might be introduced.
        """
        return self.source.getMajor()
pytact.graph_api_capnp.DataVersion.schema = _Schema()._init(getDataVersion_Schema()).as_struct()

cdef class Dataset_Reader:
    """
    Every file in the dataset contains a single message of type `Dataset`. Every file corresponds to
    a Coq source file, and contains a representation of all definitions that have existed at any point
    throughout the compilation of the source file.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asDataset()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Dataset_Reader source, object root):
        cdef Dataset_Reader wrapper = Dataset_Reader.__new__(Dataset_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def dependencies(self):
        """
        The graph contained in a file may reference nodes from the graph of other files. This field maps
        a `DepIndex` into the a file that contains that particular node.
        The first file in this list is always the current file. It is guaranteed that no cycles exist in
        the dependency relation between files induced by this field (except for the self-reference of the file).
        """
        return String_List.init(self.source.getDependencies(), self.root)
    @property
    def has_dependencies(self):
        return self.source.hasDependencies()

    @property
    def graph(self):
        return Graph_Reader.init(self.source.getGraph(), self.root)
    @property
    def has_graph(self):
        return self.source.hasGraph()

    @property
    def representative(self):
        """
        The entry point of the global context of definitions that are available when this file is 'Required' by
        another file. The full global context can be obtained by following the `previous` node of definitions.
        If the compilation unit does not contain any 'super'-global definitions this is set to `len(graph.nodes)`
        """
        return self.source.getRepresentative()

    @property
    def definitions(self):
        """
        All of the definitions present in the graph.
        Note that some of these nodes may not be part of the 'super-global' context that is reachable using the
        `representative` field as an entry point. The reason is that the global context is a forest (list of tree's)
        and the 'super-global' context is only the main spine of this forest.
        """
        return Uint32_List.init(self.source.getDefinitions(), self.root)
    @property
    def has_definitions(self):
        return self.source.hasDefinitions()

    @property
    def module_name(self):
        """
        The name of the module defined by this file.
        """
        temp = self.source.getModuleName()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_module_name(self):
        return self.source.hasModuleName()

    @property
    def data_version(self):
        """
        The version of the data stored in this dataset.
        """
        return DataVersion_Reader.init(self.source.getDataVersion(), self.root)
    @property
    def has_data_version(self):
        return self.source.hasDataVersion()
pytact.graph_api_capnp.Dataset.schema = _Schema()._init(getDataset_Schema()).as_struct()

cdef class Exception_Reader:
    """
    A list of things that can go wrong during reinforcement learning.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asException()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_Exception_Reader source, object root):
        cdef Exception_Reader wrapper = Exception_Reader.__new__(Exception_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
    @property
    def which(self):
        return Exception_Which(self.source.which())
    @property
    def which_raw(self):
        return self.source.which()

    @property
    def no_such_tactic(self):
        return self.source.getNoSuchTactic()
    @property
    def is_no_such_tactic(self):
        return self.source.isNoSuchTactic()

    @property
    def mismatched_arguments(self):
        return self.source.getMismatchedArguments()
    @property
    def is_mismatched_arguments(self):
        return self.source.isMismatchedArguments()

    @property
    def parse_error(self):
        return self.source.getParseError()
    @property
    def is_parse_error(self):
        return self.source.isParseError()

    @property
    def illegal_argument(self):
        return self.source.getIllegalArgument()
    @property
    def is_illegal_argument(self):
        return self.source.isIllegalArgument()
pytact.graph_api_capnp.Exception.schema = _Schema()._init(getException_Schema()).as_struct()

cdef class ExecutionResult_Reader:
    """
    The result of executing a tactic on a proof state.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asExecutionResult()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_ExecutionResult_Reader source, object root):
        cdef ExecutionResult_Reader wrapper = ExecutionResult_Reader.__new__(ExecutionResult_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
    @property
    def which(self):
        return ExecutionResult_Which(self.source.which())
    @property
    def which_raw(self):
        return self.source.which()

    @property
    def failure(self):
        """
        The tactic execution failed. This is not an error condition, but rather the natural failure of the tactic.
        """
        return self.source.getFailure()
    @property
    def is_failure(self):
        return self.source.isFailure()

    @property
    def complete(self):
        """
        The proof has been completed.
        """
        return self.source.getComplete()
    @property
    def is_complete(self):
        return self.source.isComplete()

    @property
    def new_state(self):
        """
        The tactic ran successfully and produced a new proof state.
        """
        return ExecutionResult_NewState_Reader.init(self.source.getNewState(), self.root)
    @property
    def is_new_state(self):
        return self.source.isNewState()

    @property
    def protocol_error(self):
        """
        Indicates a programmer error.
        """
        return Exception_Reader.init(self.source.getProtocolError(), self.root)
    @property
    def is_protocol_error(self):
        return self.source.isProtocolError()
    @property
    def has_protocol_error(self):
        return self.source.hasProtocolError()
pytact.graph_api_capnp.ExecutionResult.schema = _Schema()._init(getExecutionResult_Schema()).as_struct()

cdef class ExecutionResult_NewState_Reader:
    """
    The tactic ran successfully and produced a new proof state.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asExecutionResult_NewState()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_ExecutionResult_NewState_Reader source, object root):
        cdef ExecutionResult_NewState_Reader wrapper = ExecutionResult_NewState_Reader.__new__(ExecutionResult_NewState_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def graph(self):
        return Graph_Reader.init(self.source.getGraph(), self.root)
    @property
    def has_graph(self):
        return self.source.hasGraph()

    @property
    def state(self):
        return ProofState_Reader.init(self.source.getState(), self.root)
    @property
    def has_state(self):
        return self.source.hasState()

cdef class PredictionProtocol_Reader:
    """
    This protocol works by exchanging raw messages over a socket. The protocol is fully linear.
    Coq sends a `Request` and waits for a corresponding `Response` message. A message of a given
    type should be responded with using the obviously corresponding response type.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_PredictionProtocol_Reader source, object root):
        cdef PredictionProtocol_Reader wrapper = PredictionProtocol_Reader.__new__(PredictionProtocol_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
pytact.graph_api_capnp.PredictionProtocol.schema = _Schema()._init(getPredictionProtocol_Schema()).as_struct()

cdef class PredictionProtocol_Request_Reader:

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_PredictionProtocol_Request_Reader source, object root):
        cdef PredictionProtocol_Request_Reader wrapper = PredictionProtocol_Request_Reader.__new__(PredictionProtocol_Request_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
    @property
    def which(self):
        return PredictionProtocol_Request_Which(self.source.which())
    @property
    def which_raw(self):
        return self.source.which()

    @property
    def initialize(self):
        """
        Start a context for making tactical predictions for proof search. The context includes the tactics
        that are currently available, the definitions that are available.
        """
        return PredictionProtocol_Request_Initialize_Reader.init(self.source.getInitialize(), self.root)
    @property
    def is_initialize(self):
        return self.source.isInitialize()

    @property
    def predict(self):
        """
        Request a list of tactic predictions given the graph of a proof state.
        """
        return PredictionProtocol_Request_Predict_Reader.init(self.source.getPredict(), self.root)
    @property
    def is_predict(self):
        return self.source.isPredict()

    @property
    def synchronize(self):
        """
        Coq uses this message to synchronize the state of the protocol when exceptions have occurred.
        The contract is that the given integer needs to be echo'd back verbatim.
        """
        return self.source.getSynchronize()
    @property
    def is_synchronize(self):
        return self.source.isSynchronize()

    @property
    def check_alignment(self):
        """
        Request for the server to align the given tactics and definition to it's internal knowledge
        and report back any tactics and definitions that were not found
        """
        return self.source.getCheckAlignment()
    @property
    def is_check_alignment(self):
        return self.source.isCheckAlignment()
pytact.graph_api_capnp.PredictionProtocol.Request.schema = _Schema()._init(getPredictionProtocol_Request_Schema()).as_struct()

cdef class PredictionProtocol_Request_Initialize_Reader:
    """
    Start a context for making tactical predictions for proof search. The context includes the tactics
    that are currently available, the definitions that are available.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request_Initialize()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_PredictionProtocol_Request_Initialize_Reader source, object root):
        cdef PredictionProtocol_Request_Initialize_Reader wrapper = PredictionProtocol_Request_Initialize_Reader.__new__(PredictionProtocol_Request_Initialize_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def data_version(self):
        """
        The version number this message and subsequent messages is compatible with.
        """
        return DataVersion_Reader.init(self.source.getDataVersion(), self.root)
    @property
    def has_data_version(self):
        return self.source.hasDataVersion()

    @property
    def stack_size(self):
        """
        Initialize messages can depend on previously sent initialize messages. These messages form a stack.
        This field indicates the size of the stack that this message depends on (if a message has no
        dependencies, this field is zero). If the current stack is larger than the stack size of this message,
        then the items excess items on the stack should be discarded before adding the current message to the
        stack. It is guaranteed that the excess items will never be referenced again.

        The dependency indices of nodes contained in the graphs of the stack are indexed as follows:
        A dependency index `i` found in the graph at stack height `j` is resolved to the graph at stack height
        `j - i`. (This scheme keeps the guarantee that dependency index `0` always points to the current graph.)
        """
        return self.source.getStackSize()

    @property
    def tactics(self):
        """
        A list of tactics that Coq currently knows about.
        """
        return AbstractTactic_Reader_List.init(self.source.getTactics(), self.root)
    @property
    def has_tactics(self):
        return self.source.hasTactics()

    @property
    def graph(self):
        """
        The graph of this message. See `stackSize` on how to resolve dependency indexes in this graph.
        """
        return Graph_Reader.init(self.source.getGraph(), self.root)
    @property
    def has_graph(self):
        return self.source.hasGraph()

    @property
    def log_annotation(self):
        """
        An annotation containing file and line information on where Coq is currently processing.
        """
        temp = self.source.getLogAnnotation()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_log_annotation(self):
        return self.source.hasLogAnnotation()

    @property
    def representative(self):
        """
        Points to the last definition of the global context. All other definitions can be accessed by following
        the `Definition.previous` chain starting from this definition. If the global context is empty
        this is equal to `len(graph.nodes)`.
        """
        return self.source.getRepresentative()

cdef class PredictionProtocol_Request_Predict_Reader:
    """
    Request a list of tactic predictions given the graph of a proof state.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Request_Predict()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_PredictionProtocol_Request_Predict_Reader source, object root):
        cdef PredictionProtocol_Request_Predict_Reader wrapper = PredictionProtocol_Request_Predict_Reader.__new__(PredictionProtocol_Request_Predict_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def graph(self):
        """
        The graph may reference definitions present in the stack of initialize messages. See
        `initialize.stackSize` on how dependency indices can be resolved to a graph in the stack.
        """
        return Graph_Reader.init(self.source.getGraph(), self.root)
    @property
    def has_graph(self):
        return self.source.hasGraph()

    @property
    def state(self):
        """
        The proof state for which a prediction is requested.
        """
        return ProofState_Reader.init(self.source.getState(), self.root)
    @property
    def has_state(self):
        return self.source.hasState()

cdef class PredictionProtocol_Prediction_Reader:

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Prediction()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_PredictionProtocol_Prediction_Reader source, object root):
        cdef PredictionProtocol_Prediction_Reader wrapper = PredictionProtocol_Prediction_Reader.__new__(PredictionProtocol_Prediction_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def tactic(self):
        return Tactic_Reader.init(self.source.getTactic(), self.root)
    @property
    def has_tactic(self):
        return self.source.hasTactic()

    @property
    def arguments(self):
        return Argument_Reader_List.init(self.source.getArguments(), self.root)
    @property
    def has_arguments(self):
        return self.source.hasArguments()

    @property
    def confidence(self):
        return self.source.getConfidence()
pytact.graph_api_capnp.PredictionProtocol.Prediction.schema = _Schema()._init(getPredictionProtocol_Prediction_Schema()).as_struct()

cdef class PredictionProtocol_TextPrediction_Reader:

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_TextPrediction()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_PredictionProtocol_TextPrediction_Reader source, object root):
        cdef PredictionProtocol_TextPrediction_Reader wrapper = PredictionProtocol_TextPrediction_Reader.__new__(PredictionProtocol_TextPrediction_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def tactic_text(self):
        temp = self.source.getTacticText()
        return (<char*>temp.begin())[:temp.size()]
    @property
    def has_tactic_text(self):
        return self.source.hasTacticText()

    @property
    def confidence(self):
        return self.source.getConfidence()
pytact.graph_api_capnp.PredictionProtocol.TextPrediction.schema = _Schema()._init(getPredictionProtocol_TextPrediction_Schema()).as_struct()

cdef class PredictionProtocol_Response_Reader:
    """
    See Request for documentation.
    """

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Response()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_PredictionProtocol_Response_Reader source, object root):
        cdef PredictionProtocol_Response_Reader wrapper = PredictionProtocol_Response_Reader.__new__(PredictionProtocol_Response_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
    @property
    def which(self):
        return PredictionProtocol_Response_Which(self.source.which())
    @property
    def which_raw(self):
        return self.source.which()

    @property
    def initialized(self):
        return self.source.getInitialized()
    @property
    def is_initialized(self):
        return self.source.isInitialized()

    @property
    def prediction(self):
        return PredictionProtocol_Prediction_Reader_List.init(self.source.getPrediction(), self.root)
    @property
    def is_prediction(self):
        return self.source.isPrediction()
    @property
    def has_prediction(self):
        return self.source.hasPrediction()

    @property
    def text_prediction(self):
        """
        Output is a list of predictions with a confidence. The list is expected to be
        sorted by decreasing confidence.
        """
        return PredictionProtocol_TextPrediction_Reader_List.init(self.source.getTextPrediction(), self.root)
    @property
    def is_text_prediction(self):
        return self.source.isTextPrediction()
    @property
    def has_text_prediction(self):
        return self.source.hasTextPrediction()

    @property
    def synchronized(self):
        return self.source.getSynchronized()
    @property
    def is_synchronized(self):
        return self.source.isSynchronized()

    @property
    def alignment(self):
        return PredictionProtocol_Response_Alignment_Reader.init(self.source.getAlignment(), self.root)
    @property
    def is_alignment(self):
        return self.source.isAlignment()
pytact.graph_api_capnp.PredictionProtocol.Response.schema = _Schema()._init(getPredictionProtocol_Response_Schema()).as_struct()

cdef class PredictionProtocol_Response_Alignment_Reader:

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asPredictionProtocol_Response_Alignment()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_PredictionProtocol_Response_Alignment_Reader source, object root):
        cdef PredictionProtocol_Response_Alignment_Reader wrapper = PredictionProtocol_Response_Alignment_Reader.__new__(PredictionProtocol_Response_Alignment_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    @property
    def unaligned_tactics(self):
        return Int64_List.init(self.source.getUnalignedTactics(), self.root)
    @property
    def has_unaligned_tactics(self):
        return self.source.hasUnalignedTactics()

    @property
    def unaligned_definitions(self):
        return Node_Reader_List.init(self.source.getUnalignedDefinitions(), self.root)
    @property
    def has_unaligned_definitions(self):
        return self.source.hasUnalignedDefinitions()

cdef class ConflatableEdges_Reader:

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).asConflatableEdges()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_ConflatableEdges_Reader source, object root):
        cdef ConflatableEdges_Reader wrapper = ConflatableEdges_Reader.__new__(ConflatableEdges_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)
pytact.graph_api_capnp.ConflatableEdges.schema = _Schema()._init(getConflatableEdges_Schema()).as_struct()


cdef class Uint16_List:

    @staticmethod
    cdef init(C_Uint16_List source, object root):
        cdef Uint16_List wrapper = Uint16_List.__new__(Uint16_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return source[index]

    def __len__(self):
        return self.source.size()

cdef class Uint32_List:

    @staticmethod
    cdef init(C_Uint32_List source, object root):
        cdef Uint32_List wrapper = Uint32_List.__new__(Uint32_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return source[index]

    def __len__(self):
        return self.source.size()

cdef class Uint64_List:

    @staticmethod
    cdef init(C_Uint64_List source, object root):
        cdef Uint64_List wrapper = Uint64_List.__new__(Uint64_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return source[index]

    def __len__(self):
        return self.source.size()

cdef class Int64_List:

    @staticmethod
    cdef init(C_Int64_List source, object root):
        cdef Int64_List wrapper = Int64_List.__new__(Int64_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return source[index]

    def __len__(self):
        return self.source.size()

cdef class String_List:

    @staticmethod
    cdef init(C_String_List source, object root):
        cdef String_List wrapper = String_List.__new__(String_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        temp = source[index]
        return (<char*>temp.begin())[:temp.size()]

    def __len__(self):
        return self.source.size()
cdef class PredictionProtocol_Prediction_Reader_List:

    @staticmethod
    cdef init(C_PredictionProtocol_Prediction_Reader_List source, object root):
        cdef PredictionProtocol_Prediction_Reader_List wrapper = PredictionProtocol_Prediction_Reader_List.__new__(PredictionProtocol_Prediction_Reader_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return PredictionProtocol_Prediction_Reader.init(source[index], self.root)

    def __len__(self):
        return self.source.size()
cdef class Graph_Node_Reader_List:

    @staticmethod
    cdef init(C_Graph_Node_Reader_List source, object root):
        cdef Graph_Node_Reader_List wrapper = Graph_Node_Reader_List.__new__(Graph_Node_Reader_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return Graph_Node_Reader.init(source[index], self.root)

    def __len__(self):
        return self.source.size()
cdef class AbstractTactic_Reader_List:

    @staticmethod
    cdef init(C_AbstractTactic_Reader_List source, object root):
        cdef AbstractTactic_Reader_List wrapper = AbstractTactic_Reader_List.__new__(AbstractTactic_Reader_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return AbstractTactic_Reader.init(source[index], self.root)

    def __len__(self):
        return self.source.size()
cdef class ProofStep_Reader_List:

    @staticmethod
    cdef init(C_ProofStep_Reader_List source, object root):
        cdef ProofStep_Reader_List wrapper = ProofStep_Reader_List.__new__(ProofStep_Reader_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return ProofStep_Reader.init(source[index], self.root)

    def __len__(self):
        return self.source.size()
cdef class PredictionProtocol_TextPrediction_Reader_List:

    @staticmethod
    cdef init(C_PredictionProtocol_TextPrediction_Reader_List source, object root):
        cdef PredictionProtocol_TextPrediction_Reader_List wrapper = PredictionProtocol_TextPrediction_Reader_List.__new__(PredictionProtocol_TextPrediction_Reader_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return PredictionProtocol_TextPrediction_Reader.init(source[index], self.root)

    def __len__(self):
        return self.source.size()
cdef class ProofState_Reader_List:

    @staticmethod
    cdef init(C_ProofState_Reader_List source, object root):
        cdef ProofState_Reader_List wrapper = ProofState_Reader_List.__new__(ProofState_Reader_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return ProofState_Reader.init(source[index], self.root)

    def __len__(self):
        return self.source.size()
cdef class Graph_EdgeTarget_Reader_List:

    @staticmethod
    cdef init(C_Graph_EdgeTarget_Reader_List source, object root):
        cdef Graph_EdgeTarget_Reader_List wrapper = Graph_EdgeTarget_Reader_List.__new__(Graph_EdgeTarget_Reader_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return Graph_EdgeTarget_Reader.init(source[index], self.root)

    def __len__(self):
        return self.source.size()
cdef class Argument_Reader_List:

    @staticmethod
    cdef init(C_Argument_Reader_List source, object root):
        cdef Argument_Reader_List wrapper = Argument_Reader_List.__new__(Argument_Reader_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return Argument_Reader.init(source[index], self.root)

    def __len__(self):
        return self.source.size()
cdef class Outcome_Reader_List:

    @staticmethod
    cdef init(C_Outcome_Reader_List source, object root):
        cdef Outcome_Reader_List wrapper = Outcome_Reader_List.__new__(Outcome_Reader_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return Outcome_Reader.init(source[index], self.root)

    def __len__(self):
        return self.source.size()
cdef class Node_Reader_List:

    @staticmethod
    cdef init(C_Node_Reader_List source, object root):
        cdef Node_Reader_List wrapper = Node_Reader_List.__new__(Node_Reader_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return Node_Reader.init(source[index], self.root)

    def __len__(self):
        return self.source.size()


import sys
import os
from pathlib import Path
import capnp
import graph_api_capnp

max_node = {}
node_total = 0

rootdir = sys.argv[1]
os.chdir(rootdir)
rootdir = Path(os.getcwd())
file_list = [f for f in rootdir.glob('**/*.bin') if f.is_file()]

for f in file_list:
    print(f)
    fin = open(f)
    g = graph_api_capnp.Graph.read_packed(fin, traversal_limit_in_words=2**64-1)
    fout = open(f.with_suffix('.repacked'), 'w+b')
    g.as_builder().write(fout)



# distutils: language = c++
# distutils: libraries = capnpc capnp capnp-rpc
# distutils: sources = pytact/graph_api.capnp.cpp
# cython: c_string_type = str
# cython: c_string_encoding = default
# cython: embedsignature = True
# cython: language_level = 3

import os
import capnp
import pytact.graph_api_capnp

{%- for c in classes %}

cdef class {{ c.cython_name }}_Reader:
    {%- if c.doc != "" %}
    """
    {{ c.doc | indent(4) }}
    """
    {%- endif %}

    def __init__(self, _DynamicStructReader dyn):
        self.source = (<C_DynamicStruct_Reader>dyn.thisptr).as{{ c.cython_name }}()
        self.root = dyn # We keep a referene to the root in order to prevent it from being garbage collected

    @staticmethod
    cdef init(C_{{ c.cython_name }}_Reader source, object root):
        cdef {{ c.cython_name }}_Reader wrapper = {{ c.cython_name }}_Reader.__new__({{ c.cython_name }}_Reader)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    @property
    def dynamic(self):
        return to_python_reader(<DynamicValue.Reader>self.source, self.root)
    def __repr__(self):
        return repr(self.dynamic)

    {%- if c.union_fields != [] %}
    @property
    def which(self):
        return {{ c.cython_name }}_Which(self.source.which())
    @property
    def which_raw(self):
        return self.source.which()
    {%- endif %}

    {%- for field in c.fields %}

    @property
    def {{ field.cython_name }}(self):
        {%- if field.doc != "" %}
        """
        {{ field.doc | indent(8) }}
        """
        {%- endif %}
        {%- if field.sort == 'native' %}
        return self.source.get{{ field.cpp_name }}()
        {%- elif field.sort == 'struct' %}
        return {{ field.cython_return }}.init(self.source.get{{ field.cpp_name }}(), self.root)
        {%- elif field.sort == 'string' %}
        temp = self.source.get{{ field.cpp_name }}()
        return (<char*>temp.begin())[:temp.size()]
        {%- elif field.sort == 'enum' %}
        return {{ field.cython_return }}(self.source.get{{ field.cpp_name }}())
        {%- else %}
        error
        {%- endif %}
    {%- if field.union == True %}
    @property
    def is_{{ field.cython_name }}(self):
        return self.source.is{{ field.cpp_name }}()
    {%- endif %}
    {%- if field.gen_has == True %}
    @property
    def has_{{ field.cython_name }}(self):
        return self.source.has{{ field.cpp_name }}()
    {%- endif %}
    {%- endfor %}

{%- if not c.is_group %}
pytact.graph_api_capnp.{{ c.pycapnp_name }}.schema = _Schema()._init(get{{ c.cython_name }}_Schema()).as_struct()
{%- endif %}
{%- endfor %}


cdef class Uint16_List:

    @staticmethod
    cdef init(C_Uint16_List source, object root):
        cdef Uint16_List wrapper = Uint16_List.__new__(Uint16_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return source[index]

    def __len__(self):
        return self.source.size()

cdef class Uint32_List:

    @staticmethod
    cdef init(C_Uint32_List source, object root):
        cdef Uint32_List wrapper = Uint32_List.__new__(Uint32_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return source[index]

    def __len__(self):
        return self.source.size()

cdef class Uint64_List:

    @staticmethod
    cdef init(C_Uint64_List source, object root):
        cdef Uint64_List wrapper = Uint64_List.__new__(Uint64_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return source[index]

    def __len__(self):
        return self.source.size()

cdef class Int64_List:

    @staticmethod
    cdef init(C_Int64_List source, object root):
        cdef Int64_List wrapper = Int64_List.__new__(Int64_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return source[index]

    def __len__(self):
        return self.source.size()

cdef class String_List:

    @staticmethod
    cdef init(C_String_List source, object root):
        cdef String_List wrapper = String_List.__new__(String_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        temp = source[index]
        return (<char*>temp.begin())[:temp.size()]

    def __len__(self):
        return self.source.size()

{%- for c in classes_lists %}
cdef class {{ c.cython_name }}_List:

    @staticmethod
    cdef init(C_{{ c.cython_name }}_List source, object root):
        cdef {{ c.cython_name }}_List wrapper = {{ c.cython_name }}_List.__new__({{ c.cython_name }}_List)
        wrapper.source = source
        wrapper.root = root
        return wrapper

    def __getitem__(self, uint index):
        source = self.source
        if index >= source.size():
            raise IndexError('Out of bounds')
        return {{ c.cython_name }}.init(source[index], self.root)

    def __len__(self):
        return self.source.size()
{%- endfor %}



from dataclasses import dataclass, field
import os
import graphviz
from collections import defaultdict
from collections.abc import Sequence
from abc import ABC, abstractmethod
from pathlib import Path
from pytact.data_reader import Dataset, Definition, Node, ProofState, Original, Discharged, Substituted
import html
import inflection
import capnp
import pytact.graph_api_capnp as graph_api_capnp
import pytact.graph_api_capnp_cython as apic

class UrlMaker(ABC):

    @abstractmethod
    def definition(self, fname: Path, defid: int) -> str:
        pass
    @abstractmethod
    def proof(self, fname: Path, defid: int) -> str:
        pass
    @abstractmethod
    def outcome(self, fname: Path, defid: int, stepi: int, outcomei: int) -> str:
        pass
    @abstractmethod
    def global_context(self, fname: Path) -> str:
        pass
    @abstractmethod
    def folder(self, path: Path) -> str:
        pass
    @abstractmethod
    def root_folder(self) -> str:
        pass

@dataclass
class Settings:
    no_defaults: bool = False # Should we use default settings?
    ignore_edges: list[int] = field(default_factory=lambda: [])
    unshare_nodes: list[int] = field(default_factory=lambda: [])
    show_trivial_evar_substs: bool = False
    hide_proof_terms: bool = False
    show_edge_labels: bool = False
    order_edges: bool = False
    concentrate_edges: bool = False
    show_non_anonymized_tactics: bool = False

    def __post_init__(self):
        if not self.no_defaults:
            self.ignore_edges = [graph_api_capnp.EdgeClassification.schema.enumerants['constOpaqueDef']]
            label = graph_api_capnp.Graph.Node.Label
            self.unshare_nodes = [label.definition.value, label.sortSProp.value,
                                  label.sortProp.value, label.sortSet.value, label.sortType.value]

@dataclass
class GraphVisualizationData:
    data: dict[Path, Dataset]
    trans_deps: dict[Path, set[Path]] = field(init=False)
    graphid2path: list[Path] = field(init=False)

    def __post_init__(self):
        self.trans_deps = transitive_closure({d.filename: list(d.dependencies)
                                              for d in self.data.values()})
        self.graphid2path = [d.filename for d in sorted(self.data.values(), key=lambda d: d.graph)]

@dataclass
class GraphVisualizationOutput:
    svg: bytes
    location: list[tuple[str, str]] # tuple[Name, URL]
    active_location: int
    text: list[str] = field(default_factory=lambda: [])
    popups: list[tuple[str, str]] = field(default_factory=lambda: []) # DOM id, text

def node_label_map(node: Node) -> tuple[str, str, str]:
    enum = apic.Graph_Node_Label_Which
    label = node.label
    if d := node.definition:
        name = d.name
        return (
            'box', name.split('.')[-1],
            f"{inflection.camelize(node.label.definition.which.name.lower())} {d.name}"
        )
    match label.which:
        case enum.SORT_PROP:
            return 'ellipse', 'Prop', 'SortProp'
        case enum.SORT_S_PROP:
            return 'ellipse', 'SProp', 'SortSProp'
        case enum.SORT_SET:
            return 'ellipse', 'Set', 'SortSet'
        case enum.SORT_TYPE:
            return 'ellipse', 'Type', 'SortType'
        case enum.REL:
            return 'circle', '↑', 'rel'
        case enum.PROD:
            return 'circle', '∀', 'prod'
        case enum.LAMBDA:
            return 'circle', 'λ', 'lambda'
        case enum.LET_IN:
            return 'ellipse', 'let', 'LetIn'
        case enum.APP:
            return 'circle', '@', 'app'
        case enum.CASE_BRANCH:
            return 'ellipse', 'branch', 'CaseBranch'
        case _:
            name = inflection.camelize(label.which.name.lower())
            return 'ellipse', name, name

def truncate_string(data, maximum):
    return data[:(maximum-2)] + '..' if len(data) > maximum else data

def make_label(context, name):
    name_split = name.split('.')
    common = os.path.commonprefix([name_split, context.split('.')])
    return '.'.join(name_split[len(common):])

def graphviz_escape(s):
    return s.replace('\\', '\\\\')

def make_tooltip(d):
    return f"{inflection.camelize(d.node.label.definition.which.name.lower())} {d.name}"

def render_proof_state_text(ps: ProofState):
    return ('<br>'.join(ps.context_text) +
            '<br>----------------------<br>' + ps.conclusion_text +
            '<br><br>Raw: ' + ps.text)

class GraphVisualizator:
    def __init__(self, data: GraphVisualizationData, url_maker: UrlMaker, settings: Settings = Settings()):
        self.data = data.data
        self.trans_deps = data.trans_deps
        self.graphid2path = data.graphid2path
        self.url_maker = url_maker
        self.settings = settings
        self.node_counter = 0


        arrow_heads = [ "dot", "inv", "odot", "invdot", "invodot" ]
        edge_arrow_map = {}
        for group in graph_api_capnp.groupedEdges:
            for i, sort in enumerate(group.conflatable):
                edge_arrow_map[sort.raw] = arrow_heads[i]
        self.edge_arrow_map = edge_arrow_map

    def url_for_path(self, r: Path):
        if r in self.data:
            return r.with_suffix('').parts[-1], self.url_maker.global_context(r)
        elif len(r.parts) > 0:
            return r.parts[-1], self.url_maker.folder(r)
        else:
            return 'dataset', self.url_maker.root_folder()

    def path2location(self, path: Path):
        return [self.url_for_path(parent) for parent in list(reversed(path.parents)) + [path]]

    def dot_apply_style(self, dot):
        dot.attr('node', fontsize='11', fontname="Helvetica", margin='0', height='0.3',
                 style="rounded, filled", penwidth="0")
        dot.attr('edge', fontsize='11', fontname="Helvetica", arrowsize='0.5', penwidth="0.6")
        dot.attr('graph', fontsize='11', fontname="Helvetica", penwidth="0.6", ranksep='0.3')
        if self.settings.order_edges:
            dot.attr('graph', ordering='out')
        if self.settings.concentrate_edges:
            dot.attr('graph', concentrate='true')


    def render_node(self, dot, node: Node, shape: str, label: str, id: str | None = None,
                    tooltip: str | None = None):
        if not id:
            id = str(node)
        if not tooltip:
            tooltip = label
        url = None
        if node.definition:
            url = self.url_maker.definition(self.graphid2path[node.graph], node.nodeid)
        dot.node(id, label, URL = url, shape = shape, tooltip = tooltip)
        return id

    def render_file_node(self, dot, f: Path):
        label = f"File: {f.with_suffix('')}"
        node_id = f"file-{f}"
        dot.node(node_id, label, URL = self.url_maker.global_context(f), shape='box')
        return node_id

    def global_context(self, fname: Path):
        dot = graphviz.Digraph(format='svg')
        self.dot_apply_style(dot)
        dot.attr('graph', compound="true")

        dataset = self.data[fname]
        representative = dataset.representative
        module_name = dataset.module_name

        def render_def(dot2, d: Definition):
            label = make_label(module_name, d.name)
            if representative and representative.node == d.node:
                label = "Representative: " + label
            tooltip = make_tooltip(d)
            match d.status:
                case Original():
                    id = self.render_node(dot2, d.node, 'box', label, tooltip=tooltip)
                case Discharged(target):
                    id = self.render_node(dot2, d.node, 'box', label, tooltip=tooltip)
                    dot.edge(id, repr(target.node),
                                arrowtail="inv", dir="both", constraint="false", style="dashed")
                case Substituted(target):
                    if d.node.graph == target.node.graph:
                        id = self.render_node(dot2, d.node, 'box', label, tooltip=tooltip)
                        dot.edge(id, str(target.node),
                                    arrowtail="odot", dir="both", constraint="false", style="dashed")
                    else:
                        with dot2.subgraph() as dot3:
                            dot3.attr(rank='same')
                            id = self.render_node(dot3, d.node, 'box', label, tooltip=tooltip)
                            id2 = self.render_node(dot3, target.node, 'box',
                                                   make_label(module_name, target.name),
                                                   tooltip=make_tooltip(target))
                            dot.edge(id, id2,
                                      arrowtail="odot", dir="both", constraint="false", style="dashed")

        for cluster in dataset.clustered_definitions():

            start = str(cluster[0].node)
            ltail = None
            if len(cluster) == 1:
                render_def(dot, cluster[0])
            else:
                ltail = "cluster_"+start
                with dot.subgraph(name=ltail) as dot2:
                    dot2.attr('graph', style='rounded')
                    last = None
                    for d in cluster:
                        render_def(dot2, d)
                        if last:
                            dot2.edge(str(last.node), str(d.node), style="invis")
                        last = d

            if prev := cluster[-1].previous:
                target = str(prev.node)
                lhead = None
                if len(list(prev.cluster)) > 1:
                    lhead = "cluster_"+target
                dot.edge(str(cluster[-1].node), target, lhead = lhead, ltail = ltail)
            for fi in cluster[-1].external_previous:
                fid = self.render_file_node(dot, self.graphid2path[fi.node.graph])
                dot.edge(str(cluster[-1].node), fid, ltail = ltail)

        location = self.path2location(fname)
        return GraphVisualizationOutput(dot.pipe(), location, len(location) - 1)

    def visualize_term(self, dot, start: Node, depth, depth_ignore: set[Node] = set(),
                       maxNodes=100, seen: dict[str, str]|None=None,
                       node_label_map=node_label_map,
                       prefix='', before_prefix='', proof_state_prefix: dict[int, str] = {}
                       ) -> str:
        if seen == None:
            seen = {}
        nodes_left = maxNodes
        def recurse(node: Node, depth, context_prefix):
            nonlocal seen
            nonlocal nodes_left

            enum = graph_api_capnp.Graph.Node.Label
            match node.label.which:
                case enum.proofState:
                    node_prefix = context_prefix
                case enum.contextAssum:
                    node_prefix = context_prefix
                case enum.contextDef:
                    node_prefix = context_prefix
                case enum.evarSubst:
                    node_prefix = prefix + context_prefix
                case _:
                    node_prefix = prefix
            id = node_prefix + str(node)
            if id in seen:
                return seen[id]
            dot_id = f"c{self.node_counter}-{id}"
            seen[id] = dot_id
            nodes_left -= 1
            self.node_counter += 1
            if nodes_left < 0:
                id = 'trunc' + str(self.node_counter)
                dot.node(id, 'truncated')
                return id

            shape, label, tooltip = node_label_map(node)
            self.render_node(dot, node, shape, label, id=dot_id, tooltip=tooltip)
            if node.definition and not node in depth_ignore:
                depth -= 1
            if depth >= 0:
                if node.label.which == graph_api_capnp.Graph.Node.Label.evar:
                    # Find the evar-id
                    evarid = [c.label.proof_state.value for _, c in node.children
                              if c.label.which == graph_api_capnp.Graph.Node.Label.proofState][0]
                    print(evarid)
                    context_prefix = proof_state_prefix.get(evarid, context_prefix)

                for edge, child in node.children:
                    if edge in self.settings.ignore_edges:
                        continue
                    if child.label.which == graph_api_capnp.Graph.Node.Label.evarSubst:
                        substs = [s for _, s in child.children]
                        if not self.settings.show_trivial_evar_substs and substs[0] == substs[1]:
                            continue
                    cid = recurse(child, depth,
                                    before_prefix if edge == graph_api_capnp.EdgeClassification.evarSubstTerm
                                    else context_prefix)
                    edge_name = inflection.camelize(apic.EdgeClassification(edge).name.lower())
                    if self.settings.show_edge_labels:
                        label = edge_name
                    else:
                        label = ""
                    dot.edge(dot_id, cid, label=label, tooltip=edge_name, labeltooltip=edge_name,
                                arrowtail=self.edge_arrow_map[edge], dir="both")
            if node.label.which_raw in self.settings.unshare_nodes:
                del seen[id]
            return dot_id
        id = recurse(start, depth, before_prefix)
        return id

    def definition(self, fname: Path, definition: int, maxNodes=100):
        dot = graphviz.Digraph(format='svg')
        self.dot_apply_style(dot)

        start = self.data[fname].node_by_id(definition)
        depth_ignore = set()
        if d := start.definition:
            depth_ignore = {d.node for d in start.definition.cluster}

        self.visualize_term(dot, start, depth=0, depth_ignore=depth_ignore,
                            maxNodes=maxNodes)

        location = self.path2location(fname)
        ext_location = location
        label = "[not a definition]"
        text = []
        proof = []
        if d := start.definition:
            label = d.name
            text = [f"Type: {d.type_text}"]
            if term := d.term_text:
                text.append(f"Term: {term}")
            if d.proof:
                proof = [("Proof", self.url_maker.proof(fname, definition))]
        ext_location = (
            location +
            [(make_label(self.data[fname].module_name, label),
              self.url_maker.definition(fname, definition))] +
            proof)
        return GraphVisualizationOutput(dot.pipe(), ext_location, len(location), text)

    def proof(self, fname: Path, definition: int):
        node = self.data[fname].node_by_id(definition)
        d = node.definition
        if not d:
            assert False
        proof = d.proof
        if not proof:
            assert False

        dot = graphviz.Digraph(format='svg')
        self.dot_apply_style(dot)
        dot.attr('node', style="filled", fillcolor="white", penwidth="0.6")
        dot.attr('graph', ordering="out")
        surrogates = set()
        outcome_to_id = {}
        for i, step in enumerate(proof):
            for j, outcome in enumerate(step.outcomes):
                id = str(outcome.before.id)
                while id in surrogates:
                    id = id + '-s'
                surrogates.add(id)
                outcome_to_id[(i, j)] = id
        for i, step in enumerate(proof):
            with dot.subgraph(name='cluster_' + str(i)) as dot2:
                dot2.attr('graph', labelloc="b", style="rounded")
                if tactic := step.tactic:
                    if self.settings.show_non_anonymized_tactics:
                        tactic_text = tactic.text_non_anonymous
                    else:
                        tactic_text = tactic.text
                else:
                    tactic_text = 'unknown'
                dot2.attr(label=tactic_text)
                for j, outcome in enumerate(step.outcomes):
                    before_id = outcome_to_id[(i, j)]
                    dot2.node(before_id, label='⬤', shape='circle', fontsize="7", height="0.25pt",
                              URL = self.url_maker.outcome(fname, definition, i, j))
                    for after in outcome.after:
                        if outcome.before.id == after.id:
                            after_id = before_id + '-s'
                            style = 'dashed'
                        else:
                            after_id = str(after.id)
                            style = 'solid'
                        dot.edge(before_id, after_id, style=style)
                    if not outcome.after:
                        qedid = str('qed-'+str(i)+'-'+str(j))
                        dot2.node(qedid, label='', shape='point', height='0.05', fillcolor='black')
                        dot.edge(before_id, qedid)

        location = (self.path2location(fname) +
                    [(make_label(self.data[fname].module_name, d.name),
                      self.url_maker.definition(fname, definition)),
                     ("Proof", self.url_maker.proof(fname, definition))])
        return GraphVisualizationOutput(dot.pipe(), location, len(location) - 1)

    def outcome(self, fname: Path, definition: int, stepi: int, outcomei: int,
                      depth = 0, maxNodes=100):
        node = self.data[fname].node_by_id(definition)
        d = node.definition
        if not d:
            assert False
        proof = d.proof
        if not proof:
            assert False

        dot = graphviz.Digraph(format='svg')
        self.dot_apply_style(dot)

        outcome = proof[stepi].outcomes[outcomei]
        seen = {}

        def node_label_map_with_ctx_names(context: Sequence[Node],
                                          context_text: Sequence[str]):
            mapping = {n: s for n, s in zip(context, context_text)}
            def nlm(node: Node):
                enum = graph_api_capnp.Graph.Node.Label
                match node.label.which:
                    case enum.contextAssum:
                        name = graphviz_escape(mapping[node])
                        return 'ellipse', truncate_string(name, 20), f"ContextAssum {name}"
                    case enum.contextDef:
                        name = graphviz_escape(mapping[node])
                        return 'ellipse', truncate_string(name, 20), f"ContextDef {name}"
                    case _:
                        return node_label_map(node)
            return nlm

        popups = []

        with dot.subgraph(name='cluster_before') as dot2:
            ps = outcome.before
            dot2.attr('graph',
                      label=f"Before state\n{graphviz_escape(truncate_string(ps.conclusion_text, 70))}",
                      tooltip=f"Before state {graphviz_escape(ps.conclusion_text)}",
                      id='before-state')
            popups.append(('before-state', render_proof_state_text(ps)))
            prefix = 'before'
            self.visualize_term(dot2, ps.root, depth=depth, prefix=prefix, before_prefix=prefix,
                                maxNodes=maxNodes, seen=seen,
                                node_label_map=node_label_map_with_ctx_names(ps.context, ps.context_text))

        with dot.subgraph(name='cluster_tactic') as dot2:
            prefix = 'before'
            tactic_text = 'unknown'
            tactic_base_text = 'unknown'
            if t := outcome.tactic:
                tactic_text = t.text
                tactic_base_text = (t.base_text.replace('__argument_marker__', '_')
                                    .replace('private_constant_placeholder', '_'))
            dot2.attr('graph', label=f"Tactic\n{tactic_text}")
            dot2.node('tactic', label = tactic_base_text)
            for i, arg in enumerate(outcome.tactic_arguments):
                if arg is None:
                    dot2.node(f"tactic-arg{i}", label=f"arg {i}: unknown")
                else:
                    id = self.visualize_term(dot2, arg, depth=depth, prefix=prefix, before_prefix=prefix,
                                        maxNodes=maxNodes, seen=seen)
                    dot2.node(f"tactic-arg{i}", label=f"arg {i}")
                    dot2.edge(f"tactic-arg{i}", id)
                dot2.edge('tactic', f"tactic-arg{i}")


        for ai, after in enumerate(outcome.after):
            with dot.subgraph(name='cluster_after' + str(ai)) as dot2:
                dot2.attr('graph',
                          label=f"After state {ai}\n{graphviz_escape(truncate_string(after.conclusion_text, 70))}",
                          tooltip=f"After state {ai} {graphviz_escape(after.conclusion_text)}",
                          id=f'after-state{ai}')
                popups.append((f'after-state{ai}', render_proof_state_text(after)))
                prefix = f'after{ai}'
                self.visualize_term(dot2, after.root, depth=depth, prefix=prefix, before_prefix=prefix,
                                    maxNodes=maxNodes, seen=seen,
                                    node_label_map=node_label_map_with_ctx_names(after.context, after.context_text))

        if not self.settings.hide_proof_terms:
            with dot.subgraph(name='cluster_term') as dot2:
                dot2.attr('graph',
                          label=f"Proof term\n{graphviz_escape(truncate_string(outcome.term_text, 70))}",
                          tooltip=f"Proof term {graphviz_escape(outcome.term_text)}",
                          id='proof-term')
                popups.append(('proof-term', outcome.term_text))
                prefix = 'term'
                proof_state_prefix = {after.id: f'after{ai}' for ai, after in enumerate(outcome.after)}
                id = self.visualize_term(dot2, outcome.term, depth=depth, prefix=prefix, before_prefix='before',
                                    proof_state_prefix=proof_state_prefix,
                                    maxNodes=maxNodes, seen=seen)
                # Sometimes the subgraph is completely empty because the term is contained in another subgraph.
                # Therefore, we artificially add a extra root node
                dot2.node('artificial-root', 'TermRoot')
                dot2.edge('artificial-root', id)

        location = (self.path2location(fname) +
                    [(make_label(self.data[fname].module_name, d.name),
                      self.url_maker.definition(fname, definition)),
                     ("Proof", self.url_maker.proof(fname, definition)),
                     (f"Step {stepi} outcome {outcomei}",
                      self.url_maker.outcome(fname, definition, stepi, outcomei))])
        return GraphVisualizationOutput(dot.pipe(), location, len(location) - 1, [], popups)

    def folder(self, expand_path: Path) -> GraphVisualizationOutput:
        expand_parts = expand_path.parts

        dot = graphviz.Digraph(engine='dot', format='svg')
        self.dot_apply_style(dot)

        hierarchy = {'files': [], 'subdirs': {}}
        for f in self.data:
            dirs = f.parent.parts
            leaf = hierarchy
            for d in dirs:
                leaf['subdirs'].setdefault(d, {'files': [], 'subdirs': {}})
                leaf = leaf['subdirs'][d]
            leaf['files'].append(f)
        def common_length(p1, p2):
            return len(os.path.commonprefix([p1, p2]))
        def retrieve_edges(rel, h, depth: int):
            for n in h['files']:
                deps = {Path(*d.parts[:depth+1]) for d in self.trans_deps[n]
                        if common_length(d.parts, expand_parts) == depth}
                rel[Path(*n.parts[:depth+1])] |= deps
            for d in h['subdirs']:
                retrieve_edges(rel, h['subdirs'][d], depth)
            return rel
        def tunnel_hierarchy(dot, h, depth):
            if depth == len(expand_parts):
                rel = retrieve_edges(defaultdict(set), h, depth)
                (rel, repr) = transitive_reduction(rel)
                for n in rel:
                    if len(repr[n]) == 1:
                        label, url = self.url_for_path(n)
                        shape = 'box'
                    else:
                        label = '<<table border="0" cellborder="0" cellpadding="7">'
                        for r in repr[n]:
                            slabel, surl = self.url_for_path(r)
                            label += f'<tr><td href="{html.escape(surl)}">{slabel}</td></tr>'
                        label += "</table>>"
                        url = None
                        shape = 'plaintext'
                    dot.node(str(n), label, URL = url, shape = shape)
                for n, deps in rel.items():
                    for d in deps:
                        dot.edge(str(n), str(d))
            else:
                for d in h['subdirs']:
                    cur_path: Path = Path(*expand_parts[:depth]) / d
                    if depth < len(expand_parts) and d == expand_parts[depth]:
                        cluster_name = 'cluster_' + str(cur_path)
                        with dot.subgraph(name=cluster_name) as dot2:
                            dot2.attr('graph', style="filled", fillcolor="white", label=d,
                                      URL = self.url_maker.folder(Path(*expand_parts[:depth+1])))
                            tunnel_hierarchy(dot2, h['subdirs'][d], depth+1)

        with dot.subgraph(name='cluster_root') as dot2:
            dot2.attr('graph', style="filled", fillcolor="white", label='dataset',
                      URL = self.url_maker.root_folder())
            tunnel_hierarchy(dot2, hierarchy, 0)

        location = self.path2location(expand_path)
        return GraphVisualizationOutput(dot.pipe(), location, len(location) - 1)

def transitive_closure(rel):
    trans_deps = defaultdict(set)
    def calc_trans_deps(a, n):
        for d in rel[n]:
            if d not in trans_deps[a]:
                trans_deps[a].add(d)
                calc_trans_deps(a, d)
    for n in list(rel.keys()):
        calc_trans_deps(n, n)
    return trans_deps

def transitive_reduction(rel):
    """This is not a real transitive reduction (NP-hard when it needs to be a subgraph).
    This is an approximation where all strongly connected components are smashed together
    in one node.
    """
    trans_deps = transitive_closure(rel)
    repr_items = defaultdict(set)
    representative = {}
    def calc_components(n):
        if n in representative:
            return
        repr_items[n].add(n)
        representative[n] = n
        for d in trans_deps[n]:
            if n in trans_deps[d]:
                repr_items[n].add(d)
                representative[d] = n
    for n in list(rel.keys()):
        calc_components(n)
    repr_trans_rel = defaultdict(set)
    def calc_new_trans_deps(n):
        for d in trans_deps[n]:
            if n not in trans_deps[d]:
                repr_trans_rel[n].add(representative[d])
    for n in list(rel.keys()):
        calc_new_trans_deps(n)
    def calc_sparse_deps(n):
        res = set()
        for d in repr_trans_rel[n]:
            k = [d2 for d2 in repr_trans_rel[n]
                 if d != d2 and d in trans_deps[d2]]
            if not k:
                res.add(d)
        return res
    k = {n: calc_sparse_deps(n) for n in repr_items}
    return (k, repr_items)



from setuptools import setup, Extension
import subprocess
import os
from shutil import copyfile
import capnp

from Cython.Build import cythonize

def capnp2cython():

    api_file = "pytact/graph_api.capnp"
    cython_file_pyx = api_file.replace('.', '_')+'_cython.pyx'
    cython_file_pxd = api_file.replace('.', '_')+'_cython.pxd'
    template_files = ["pytact/templates/capnp_cython.pyx", "pytact/templates/capnp_cython.pxd"]

    # Poor man's make
    target_time = min([(os.path.getmtime(f) if os.path.exists(f) else 0) for f in
                        [api_file+'.c++', api_file+'.cpp', api_file+'.h', cython_file_pyx, cython_file_pxd]])
    source_time = max([os.path.getmtime(f) for f in [api_file] + template_files])
    if target_time < source_time:
        print(subprocess.check_output(["capnpc", "-oc++", api_file]))
        subprocess.check_output(["capnpc", "-o./pytact/generate_api.py", api_file])

        copyfile(api_file + '.c++', api_file + '.cpp')

    capnp_loc = os.path.abspath(os.path.join(list(capnp.__path__)[0], '../'))
    return cythonize([
        Extension(
            name = "pytact.graph_api_capnp_cython",
            sources = [cython_file_pyx],
            extra_compile_args=["-O3", "-std=c++14"],
            include_dirs=[capnp_loc]
        ),
        Extension(
            name = "pytact.data_reader",
            sources = ['pytact/data_reader.pyx'],
            extra_compile_args=["-O3", "-std=c++14"],
            include_dirs=[capnp_loc]
        )])

setup(
    ext_modules = capnp2cython()
)



"""
common  wrapping functions for python - capnp - tactician interface

provides class Proving with method run_prover that evaluates your python
proving session in coq environment over capnp communication

you need yourself to establish socket/pipe connection with capnp end of coq
and instantiate Proving object with
"""

import pkg_resources
import os
import logging
import asyncio
import functools
from typing import List
import capnp

test_filename_stdin = pkg_resources.resource_filename('pytact','tests/TestReinforceStdin.v')
test_filename_tcp  =  pkg_resources.resource_filename('pytact','tests/TestReinforceTcp.v')

def testfile(fname: str):
    return pkg_resources.resource_filename('pytact', os.path.join('tests', fname))


logger = logging.getLogger(__name__)





class Connection:
    """
    wrapper class for serving capnp  connection
    this wrapper is not specialized to a particular capnp schema
    """
    async def _reading_socket(self):
        """
        This loop reads from connected pipe and writes to local capnp client
        """
        client = self.client
        reader = self.reader
        while not reader.at_eof():
            data = await reader.read(4096)
            logger.debug("read from socket %d bytes", len(data))
            client.write(data)
            logger.debug("wrote to client %d bytes", len(data))

        logger.debug("_reading_socket ended")
        return '_reading_socket ended'


    async def _writing_socket(self):
        """
        This reads from local capnp client and writes to connected pipe
        """
        client = self.client
        writer = self.writer
        while True:
            data = await client.read(4096)
            logger.debug("read from client %d bytes", len(data))

            writer.write(data.tobytes())
            await writer.drain()
            logger.debug("wrote to socket %d bytes", len(data))

        logger.debug("_writing to socket ended")
        return '_writing_socket ended'

    def __init__(self, reader, writer):
        self.reader = reader
        self.writer = writer

        logger.debug("starting capnp client")

        import pytact.graph_api_capnp as api
        self.api = api

        self.client = capnp.TwoPartyClient()



    async def run(self, proc):
        """
        this method runs the function proc
        that defines the process of interacting with capnp connection
        on python side

        the function proc should be of async awaitable type that takes
        two arguments, proc(client, api), the types of client and api
        are like in these cals

        client = capnp.TwoPartyClient()
        api = capnp.load(api.capnp)

        and returns None when completed

        example:

        async def proc(client)
            # do your interaction with with capnp client using the api
            return

        """

        read_task = asyncio.create_task(self._reading_socket())
        write_task = asyncio.create_task(self._writing_socket())
        proc_task = asyncio.create_task(proc(self.client, self.api))

        _done, _pending = await asyncio.wait([read_task, write_task, proc_task],
                                           return_when=asyncio.FIRST_COMPLETED)

        if proc_task.done():
            logger.debug("process with capnp ended %s", repr(proc_task))
            if proc_task.exception():
                raise Exception("EXCEPTION in connection task; this is not expected, debug: %s",
                                repr(proc_task))

        else:
            if read_task.done():
                raise Exception("reading_socket ended before reinforce ended: "
                             "possibly terminated from other side")
            if write_task.done():
                raise Exception("_writing_socket ended ended before reinforce ended: "
                             "possibly connection broken")

        read_task.cancel()
        write_task.cancel()
        proc_task.cancel()
        self.writer.close()

        logger.info("serving connection finished")

class Proving(Connection):
    """
    this is a higher level wrapper over Connection class that
    provides higher level run_prover method that takes as an input
    and runs your function prover and provides
    to prover the pull object of graph_api.capnp schema

    Example:

    proving = Proving(reader, writer)
    await proving.run_prover(your_super_prover)

    where reader and writer are the read and write end of socket/pipe to
    the other end of capnp coq

    and where you need to implement your_super_prover function as you like

    See doc string of run_prover method
    """
    @staticmethod
    async def __proc(prover, client, api):
        logger.info("calling PullReinforce")
        pull = client.bootstrap().cast_as(api.PullReinforce)
        await prover(pull)

    async def run_prover(self, prover):
        """
        The method run_prover runs your function prover in coq
        evaluation session  using capnp protocol graph_api.capnp schema

        Your function prover should take one argument

        pull

        that is an object defined by graph_api.capnp schema

        You can start the  proving session in your functin prover
        by opening coq proving environment with a call to pull as follows:

        response = await pull.reinforce(lemma).a_wait

        You get available tactics with

        available = await response.available.tactics().a_wait

        You get the context and goal packed in response.result
        as specified in graph_api.capnp schema

        NOTICE: the implementation of this class depends on the particular naming
        and data structures we have chosen in graph_api.capnp schema

        This is subject to change
        """
        await self.run(functools.partial(self.__proc, prover))


def globalNode(nodeIndex):
    """
    returns capnp global node argument given local node index
    """
    return {'depIndex': 0, 'nodeIndex': nodeIndex}

def capnpLocalTactic(ident: int, local_arguments: List[int]):
    """
    returns capnpTactic argument given  ident and a list of local nodes
    """
    res = {'ident': ident}, [{'term': globalNode(nodeIndex)} for nodeIndex in local_arguments]
#           'arguments': [{'term': globalNode(nodeIndex)} for nodeIndex in local_arguments]}
    return res

def graph_api_capnp():
    return pkg_resources.resource_filename('pytact','graph_api.capnp')

async def runTactic(obj, tactic, vis=None):
    """
    wrapper on capnp runTactic with logging
    vis: pytact.graph_visualize.Visualizer
    """
    logger.info("awaiting tactic %s", repr(tactic))
    if isinstance(tactic, str):
        resp = await obj.runTextTactic(tactic).a_wait()
    else:
        resp = await obj.runTactic(*tactic).a_wait()
    result = resp.result
    logger.info('received result of type %s', result.which())
    if result.which() == 'newState':
        logger.info('new proof state: %s', result.newState.state.text)
        root = result.newState.state.root
        context = list(result.newState.state.context)
        logger.debug('with root: %d', root)
        logger.debug('with context: %s', repr(context))
    if vis is not None:
        vis.render(result)

    if result.which() == 'protocolError':
        logger.error('received protocolError %s', result.protocolError.which())
    #    raise ProtocolError  # our code is wrong

    return result



"""
TODO: check if works, fix or deprecate

This file may be useful to Vasily/Fidel/Mirek/Jason when you want
to test a reinforcement client without having to install the Opam
packages. Wasn't maintained for a while
"""


import os
import sys

import socket
import asyncio

from pathlib import Path

import capnp
capnp.remove_import_hook()

graph_api_capnp = pytact.common.graph_api_capnp()
graph_api_capnp = capnp.load(graph_api_capnp)

class ProofObjectImpl(graph_api_capnp.ProofObject.Server):

    def runTactic(self, tactic, _context):
        print("Fake Coq: run tactic")
        print(tactic)
        return {
            'newState': {
                'graph': {
                    'classifications': [],
                    'edges': []
                },
                'state': {
                    'root': 2,
                    'context': []
                },
                'obj': ProofObjectImpl()
            }
        }

class PullReinforceImpl(graph_api_capnp.PullReinforce.Server):

    def reinforce(self, lemma, _context):
        print('Fake Coq: start')
        print(lemma)
        return {
            'newState': {
                'graph': {
                    'classifications': [],
                    'edges': []
                },
                'state': {
                    'root': 2,
                    'context': []
                },
                'obj': ProofObjectImpl()
            }
        }

class MainImpl(graph_api_capnp.Main.Server):

    def initialize(self, push, _context):
        print('Fake Coq: initialize')
        print(push)
        state = {
            'newState': {
                'graph': {
                    'classifications': [],
                    'edges': []
                },
                'state': {
                    'root': 2,
                    'context': []
                },
                'obj': ProofObjectImpl()
            }
        }
        return push.reinforce(state).then(
            lambda _: setattr(_context.results, "pull", PullReinforceImpl()))

async def readloop(reader, server, write_task):
    while not reader.at_eof():
        data = await reader.read(4096)
        await server.write(data)
    write_task.cancel()


async def writeloop(writer, server):
    while True:
        data = await server.read(4096)
        writer.write(data.tobytes())


async def serve_capnp(reader, writer):
    server = capnp.TwoPartyServer(bootstrap=MainImpl())

    write_task = asyncio.create_task(writeloop(writer, server))
    coroutines = [write_task, readloop(reader, server, write_task)]

    tasks = asyncio.gather(*coroutines, return_exceptions=True)
    await tasks


async def main():
    s = socket.socket(fileno=sys.stdin.fileno())
    reader, writer = await asyncio.open_connection(sock=s)
    await serve_capnp(reader, writer)


if __name__ == '__main__':
    asyncio.run(main())
